#!/usr/bin/env python

# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>

# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                     HEALTH INPATIENT package                          #
#                 health_inpatient.py: Main module                      #
#########################################################################
from datetime import datetime
from trytond.model import ModelView, ModelSQL, fields, Unique
from trytond.transaction import Transaction
from trytond.tools import grouped_slice, reduce_ids
from sql import Literal
from trytond.pool import Pool
from trytond.pyson import Eval, Not, Bool, And, Equal, Or
from trytond.i18n import gettext
from trytond.modules.health.core import get_health_professional, \
    get_institution

import pytz

from .exceptions import (
    NoAssociatedHealthProfessional,
    DischargeReasonNeeded, DischargeBeforeAdmission,
    BedIsNotAvailable, NeedTimeZone,
    AdmissionMustBeToday, SpecialMealNeeds
)


__all__ = [
    'DietTherapeutic', 'InpatientRegistration',
    'BedTransfer', 'Appointment', 'PatientEvaluation', 'PatientData',
    'InpatientMedication', 'InpatientMedicationAdminTimes',
    'InpatientMedicationLog', 'InpatientDiet', 'InpatientMeal',
    'InpatientMealOrder', 'InpatientMealOrderItem', 'ECG',
    'PatientRounding', 'RoundingProcedure']

# Therapeutic Diet types


class DietTherapeutic (ModelSQL, ModelView):
    'Diet Therapy'
    __name__ = "gnuhealth.diet.therapeutic"

    name = fields.Char('Diet type', required=True, translate=True)
    code = fields.Char('Code', required=True)
    description = fields.Text('Indications', required=True, translate=True)

    @classmethod
    def __setup__(cls):
        super(DietTherapeutic, cls).__setup__()

        t = cls.__table__()
        cls._sql_constraints = [
            ('code_unique', Unique(t, t.code),
                'The Diet code already exists'),
        ]


class InpatientRegistration(ModelSQL, ModelView):
    'Patient admission History'
    __name__ = 'gnuhealth.inpatient.registration'

    STATES = {'readonly': Or(
        Eval('state') == 'done',
        Eval('state') == 'finished')}

    name = fields.Char('Registration Code', readonly=True)
    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient',
        required=True, states=STATES)
    admission_type = fields.Selection([
        (None, ''),
        ('routine', 'Routine'),
        ('maternity', 'Maternity'),
        ('elective', 'Elective'),
        ('urgent', 'Urgent'),
        ('emergency', 'Emergency'),
    ], 'Admission type', required=True, states=STATES)
    hospitalization_date = fields.DateTime(
        'Hospitalization date',
        required=True, states=STATES)
    discharge_date = fields.DateTime(
        'Expected Discharge Date', required=True,
        states=STATES)
    attending_physician = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Attending Physician', states=STATES)
    operating_physician = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Operating Physician', states=STATES)
    admission_reason = fields.Many2One(
        'gnuhealth.pathology',
        'Reason for Admission', help="Reason for Admission", states=STATES)
    bed = fields.Many2One(
        'gnuhealth.hospital.bed', 'Hospital Bed',
        states={
            'required': Not(Bool(Eval('name'))),
            'readonly': Or(
                Eval('state') == 'done',
                Eval('state') == 'finished',
                Bool(Eval('name')),
            )
        },
        depends=['name'])
    nursing_plan = fields.Text('Nursing Plan', states=STATES)
    medications = fields.One2Many(
        'gnuhealth.inpatient.medication', 'registration',
        'Medications', states=STATES)
    therapeutic_diets = fields.One2Many(
        'gnuhealth.inpatient.diet', 'registration',
        'Meals / Diet Program', states=STATES)

    nutrition_notes = fields.Text(
        'Nutrition notes / directions', states=STATES)
    discharge_plan = fields.Text(
        'Discharge Plan', states=STATES)
    info = fields.Text('Notes', states=STATES)
    state = fields.Selection((
        (None, ''),
        ('free', 'free'),
        ('cancelled', 'cancelled'),
        ('confirmed', 'confirmed'),
        ('hospitalized', 'hospitalized'),
        ('done', 'Discharged - needs cleaning'),
        ('finished', 'Finished'),
    ), 'Status', readonly=True)

    bed_transfers = fields.One2Many(
        'gnuhealth.bed.transfer', 'registration',
        'Transfer History', readonly=True)

    discharged_by = fields.Many2One(
        'gnuhealth.healthprofessional', 'Discharged by', readonly=True,
        help="Health Professional that discharged the patient")

    discharge_reason = fields.Selection([
        (None, ''),
        ('home', 'Home / Selfcare'),
        ('transfer', 'Transferred to another institution'),
        ('death', 'Death'),
        ('against_advice', 'Left against medical advice')],
        'Discharge Reason',
        states={'readonly': Not(Equal(Eval('state'), 'hospitalized'))},
        help="Reason for patient discharge")

    discharge_dx = fields.Many2One(
        'gnuhealth.pathology',
        'Discharge Dx', help="Code for Discharge Diagnosis",
        states={'readonly': Not(Equal(Eval('state'), 'hospitalized'))})

    institution = fields.Many2One(
        'gnuhealth.institution', 'Institution',
        readonly=True)

    puid = fields.Function(
        fields.Char('PUID', help="Person Unique Identifier"),
        'get_patient_puid', searcher="search_patient_puid")

    @staticmethod
    def default_institution():
        return get_institution()

    def get_patient_puid(self, name):
        return self.patient.party.ref

    @classmethod
    def search_patient_puid(cls, name, clause):
        res = []
        value = clause[2]
        res.append(('patient.party.ref', clause[1], value))
        return res

    @classmethod
    def __setup__(cls):
        super(InpatientRegistration, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('name_unique', Unique(t, t.name),
                'The Registration code already exists'),
        ]

        cls._buttons.update({
            'confirmed': {
                'invisible': And(
                    Not(Equal(Eval('state'), 'free')),
                    Not(Equal(Eval('state'), 'cancelled'))),
            },
            'cancel': {
                'invisible': Not(Equal(Eval('state'), 'confirmed')),
            },
            'admission': {
                'invisible': Not(Equal(Eval('state'), 'confirmed')),
            },
            'discharge': {
                'invisible': Not(Equal(Eval('state'), 'hospitalized')),
            },
            'bedclean': {
                'invisible': Not(Equal(Eval('state'), 'done')),
            },
        })

    # Method to check for availability and make the hospital bed reservation
    # Checks that there are not overlapping dates and status of the bed / room
    # is not confirmed, hospitalized or done but requiring cleaning ('done')
    @classmethod
    @ModelView.button
    def confirmed(cls, registrations):
        pool = Pool()
        Bed = pool.get('gnuhealth.hospital.bed')
        cursor = Transaction().connection.cursor()
        table = cls.__table__()

        beds = []
        for registration in registrations:
            if (registration.discharge_date.date() <
                    registration.hospitalization_date.date()):
                raise DischargeBeforeAdmission(
                    gettext('health_inpatient.msg_discharge_befor_admission'))
            cursor.execute(*table.select(
                table.id,
                where=((
                    (table.hospitalization_date <=
                     registration.hospitalization_date) &
                    (table.discharge_date >=
                     registration.hospitalization_date)
                ) |
                    ((table.hospitalization_date
                      <= registration.discharge_date)
                        & (table.discharge_date
                           >= registration.discharge_date)) |
                    ((table.hospitalization_date
                      >= registration.hospitalization_date)
                     & (table.discharge_date
                        <= registration.discharge_date)))
                & table.state.in_(['confirmed', 'hospitalized', 'done'])
                & (table.bed == registration.bed.id)))

            if cursor.fetchone():
                raise BedIsNotAvailable(
                    gettext('health_inpatient.msg_bed_is_not_available'))
            beds.append(registration.bed)
        cls.write(registrations, {'state': 'confirmed'})
        Bed.write(beds, {'state': 'reserved'})

    @classmethod
    @ModelView.button
    def discharge(cls, registrations):
        registration_id = registrations[0]
        Bed = Pool().get('gnuhealth.hospital.bed')

        signing_hp = get_health_professional()
        if not signing_hp:
            raise NoAssociatedHealthProfessional(
                gettext('health_inpatient.'
                        'msg_no_associated_health_professional'))

        cls.write(registrations, {
            'state': 'done', 'discharged_by': signing_hp})

        Bed.write([registration_id.bed], {'state': 'to_clean'})

    @classmethod
    @ModelView.button
    def bedclean(cls, registrations):
        registration_id = registrations[0]
        Bed = Pool().get('gnuhealth.hospital.bed')

        cls.write(registrations, {'state': 'finished'})

        Bed.write([registration_id.bed], {'state': 'free'})

    @classmethod
    @ModelView.button
    def cancel(cls, registrations):
        registration_id = registrations[0]
        Bed = Pool().get('gnuhealth.hospital.bed')

        cls.write(registrations, {'state': 'cancelled'})
        Bed.write([registration_id.bed], {'state': 'free'})

    @classmethod
    @ModelView.button
    def admission(cls, registrations):
        registration_id = registrations[0]
        Bed = Pool().get('gnuhealth.hospital.bed')

        Company = Pool().get('company.company')

        timezone = None
        company_id = Transaction().context.get('company')
        if company_id:
            company = Company(company_id)
            if company.timezone:
                timezone = pytz.timezone(company.timezone)

                dt = datetime.today()
                dt_local = datetime.astimezone(
                    dt.replace(tzinfo=pytz.utc), timezone)

                if (registration_id.hospitalization_date.date() !=
                        dt_local.date()):
                    raise AdmissionMustBeToday(
                        gettext('health_inpatient.'
                                'msg_admission_must_be_today'))
                else:
                    cls.write(registrations, {'state': 'hospitalized'})
                    Bed.write([registration_id.bed], {'state': 'occupied'})

            else:
                raise NeedTimeZone(
                    gettext('health_inpatient.msg_need_timezone'))

    @classmethod
    def generate_code(cls, **pattern):
        Config = Pool().get('gnuhealth.sequences')
        config = Config(1)
        sequence = config.get_multivalue(
            'inpatient_registration_sequence', **pattern)
        if sequence:
            return sequence.get()

    @classmethod
    def create(cls, vlist):
        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('name'):
                values['name'] = cls.generate_code()
        return super(InpatientRegistration, cls).create(vlist)

    @staticmethod
    def default_state():
        return 'free'

    @classmethod
    def validate(cls, registrations):
        super(InpatientRegistration, cls).validate(registrations)
        for registration in registrations:
            registration.check_discharge_context()

    def check_discharge_context(self):
        if ((not self.discharge_reason or not self.discharge_dx
            or not self.admission_reason)
                and self.state == 'done'):
            raise DischargeReasonNeeded(
                gettext('health_inpatient.msg_discharge_reason_needed'))

    # Format Registration ID : Patient : Bed
    def get_rec_name(self, name):
        if self.patient:
            return f'{self.name}:{self.bed.rec_name}:{self.patient.rec_name}'
        else:
            return self.name

    # Allow searching by the hospitalization code, patient name
    # or bed number

    @classmethod
    def search_rec_name(cls, name, clause):
        field = None
        # Search by Registration Code ID or Patient
        for field in ('name', 'patient', 'bed'):
            registrations = cls.search([(field,) + tuple(clause[1:])], limit=1)
            if registrations:
                break
        if registrations:
            return [(field,) + tuple(clause[1:])]
        return [(cls._rec_name,) + tuple(clause[1:])]


class BedTransfer(ModelSQL, ModelView):
    'Bed transfers'
    __name__ = 'gnuhealth.bed.transfer'

    registration = fields.Many2One(
        'gnuhealth.inpatient.registration',
        'Registration Code')
    transfer_date = fields.DateTime('Date')
    bed_from = fields.Many2One('gnuhealth.hospital.bed', 'From')
    bed_to = fields.Many2One('gnuhealth.hospital.bed', 'To')
    reason = fields.Char('Reason')

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to registration
        if (table_h.column_exist('name')
                and not table_h.column_exist('registration')):
            table_h.column_rename('name', 'registration')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class Appointment(ModelSQL, ModelView):
    __name__ = 'gnuhealth.appointment'

    inpatient_registration_code = fields.Many2One(
        'gnuhealth.inpatient.registration', 'Inpatient',
        help="Inpatient Registration, "
        "please enter the patient hospitalization code.")


class PatientEvaluation(ModelSQL, ModelView):
    __name__ = 'gnuhealth.patient.evaluation'

    inpatient_registration_code = fields.Many2One(
        'gnuhealth.inpatient.registration', 'IPC',
        help="Inpatient Code, "
        "please enter the patient hospitalization code")


class ECG(ModelSQL, ModelView):
    __name__ = 'gnuhealth.patient.ecg'

    inpatient_registration_code = fields.Many2One(
        'gnuhealth.inpatient.registration', 'Inpatient',
        help="Inpatient Registration, "
        "please enter the patient hospitalization code")


class PatientData(ModelSQL, ModelView):
    'Inherit patient model and add the patient status to the patient.'
    __name__ = 'gnuhealth.patient'

    patient_status = fields.Function(
        fields.Boolean(
            'Hospitalized',
            help="Show the hospitalization status of the patient"),
        getter='get_patient_status', searcher='search_patient_status')

    @classmethod
    def get_patient_status(cls, patients, name):
        cursor = Transaction().connection.cursor()
        pool = Pool()
        Registration = pool.get('gnuhealth.inpatient.registration')
        registration = Registration.__table__()

        # Will store statuses {patient: True/False, ...}
        ids = list(map(int, patients))
        result = dict.fromkeys(ids, False)

        for sub_ids in grouped_slice(ids):
            # SQL expression for relevant patient ids
            clause_ids = reduce_ids(registration.id, sub_ids)

            # Hospitalized patient ids
            query = registration.select(
                registration.patient, Literal(True),
                where=(registration.state == 'hospitalized') & clause_ids,
                group_by=registration.patient)

            # Update dictionary of patient ids with True statuses
            cursor.execute(*query)
            result.update(cursor.fetchall())

        return result

    @classmethod
    def search_patient_status(cls, name, clause):
        p = Pool().get('gnuhealth.inpatient.registration')
        table = p.__table__()
        pat = cls.__table__()
        _, operator, value = clause

        # Validate operator and value
        if operator not in ['=', '!=']:
            raise ValueError('Wrong operator: %s' % operator)
        if value is not True and value is not False:
            raise ValueError('Wrong value: %s' % value)

        # Find hospitalized patient ids
        j = pat.join(table, condition=pat.id == table.patient)
        s = j.select(pat.id, where=table.state == 'hospitalized')

        # Choose domain operator
        if (operator == '=' and value) or (operator == '!=' and not value):
            d = 'in'
        else:
            d = 'not in'

        return [('id', d, s)]


class InpatientMedication (ModelSQL, ModelView):
    'Inpatient Medication'
    __name__ = 'gnuhealth.inpatient.medication'

    registration = fields.Many2One(
        'gnuhealth.inpatient.registration',
        'Registration Code')
    medicament = fields.Many2One(
        'gnuhealth.medicament', 'Medicament',
        required=True, help='Prescribed Medicament')
    indication = fields.Many2One(
        'gnuhealth.pathology', 'Indication',
        help='Choose a disease for this medicament from the disease list. It'
        ' can be an existing disease of the patient or a prophylactic.')
    start_treatment = fields.DateTime(
        'Start',
        help='Date of start of Treatment', required=True)
    end_treatment = fields.DateTime('End', help='Date of start of Treatment')
    dose = fields.Float(
        'Dose',
        help='Amount of medication (eg, 250 mg) per dose', required=True)
    dose_unit = fields.Many2One(
        'gnuhealth.dose.unit', 'dose unit',
        required=True, help='Unit of measure for the medication to be taken')
    route = fields.Many2One(
        'gnuhealth.drug.route', 'Administration Route',
        required=True, help='Drug administration route code.')
    form = fields.Many2One(
        'gnuhealth.drug.form', 'Form', required=True,
        help='Drug form, such as tablet or gel')
    qty = fields.Integer(
        'x', required=True,
        help='Quantity of units (eg, 2 capsules) of the medicament')
    common_dosage = fields.Many2One(
        'gnuhealth.medication.dosage', 'Frequency',
        help='Common / standard dosage frequency for this medicament')
    admin_times = fields.One2Many(
        'gnuhealth.inpatient.medication.admin_time',
        'medication', "Admin times")
    log_history = fields.One2Many(
        'gnuhealth.inpatient.medication.log', 'medication',
        "Log History")
    frequency = fields.Integer(
        'Frequency',
        help='Time in between doses the patient must wait (ie, for 1 pill'
        ' each 8 hours, put here 8 and select \"hours\" in the unit field')
    frequency_unit = fields.Selection([
        (None, ''),
        ('seconds', 'seconds'),
        ('minutes', 'minutes'),
        ('hours', 'hours'),
        ('days', 'days'),
        ('weeks', 'weeks'),
        ('wr', 'when required'),
    ], 'unit', sort=False)
    frequency_prn = fields.Boolean('PRN', help='Use it as needed, pro re nata')

    is_active = fields.Boolean(
        'Active',
        help='Check if the patient is currently taking the medication')
    discontinued = fields.Boolean('Discontinued')
    course_completed = fields.Boolean('Course Completed')
    discontinued_reason = fields.Char(
        'Reason for discontinuation',
        states={
            'invisible': Not(Bool(Eval('discontinued'))),
            'required': Bool(Eval('discontinued')),
        },
        depends=['discontinued'],
        help='Short description for discontinuing the treatment')
    adverse_reaction = fields.Text(
        'Adverse Reactions',
        help='Side effects or adverse reactions that the patient experienced')

    @fields.depends('discontinued', 'course_completed')
    def on_change_with_is_active(self):
        is_active = True
        if (self.discontinued or self.course_completed):
            is_active = False
        return is_active

    @fields.depends('is_active', 'course_completed')
    def on_change_with_discontinued(self):
        return not (self.is_active or self.course_completed)

    @fields.depends('is_active', 'discontinued')
    def on_change_with_course_completed(self):
        return not (self.is_active or self.discontinued)

    @staticmethod
    def default_is_active():
        return True

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to registration
        if (table_h.column_exist('name')
                and not table_h.column_exist('registration')):
            table_h.column_rename('name', 'registration')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class InpatientMedicationAdminTimes (ModelSQL, ModelView):
    'Inpatient Medication Admin Times'
    __name__ = "gnuhealth.inpatient.medication.admin_time"

    medication = fields.Many2One(
        'gnuhealth.inpatient.medication', 'Medication')
    admin_time = fields.Time("Time")
    dose = fields.Float(
        'Dose',
        help='Amount of medication (eg, 250 mg) per dose')
    dose_unit = fields.Many2One(
        'gnuhealth.dose.unit', 'dose unit',
        help='Unit of measure for the medication to be taken')
    remarks = fields.Text(
        'Remarks',
        help='specific remarks for this dose')

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name
        if (table_h.column_exist('name')
                and not table_h.column_exist('medication')):
            table_h.column_rename('name', 'medication')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class InpatientMedicationLog (ModelSQL, ModelView):
    'Inpatient Medication Log History'
    __name__ = "gnuhealth.inpatient.medication.log"

    medication = fields.Many2One(
        'gnuhealth.inpatient.medication', 'Medication')
    admin_time = fields.DateTime("Date", readonly=True)
    health_professional = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Health Prof', readonly=True)
    dose = fields.Float(
        'Dose',
        help='Amount of medication (eg, 250 mg) per dose')
    dose_unit = fields.Many2One(
        'gnuhealth.dose.unit', 'dose unit',
        help='Unit of measure for the medication to be taken')
    remarks = fields.Text(
        'Remarks',
        help='specific remarks for this dose')

    @classmethod
    def validate(cls, records):
        super(InpatientMedicationLog, cls).validate(records)
        for record in records:
            record.check_health_professional()

    def check_health_professional(self):
        if not self.health_professional:
            raise NoAssociatedHealthProfessional(
                gettext('health.msg_no_associated_health_professional'))

    @staticmethod
    def default_health_professional():
        return get_health_professional()

    @staticmethod
    def default_admin_time():
        return datetime.now()

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to medication
        if (table_h.column_exist('name')
                and not table_h.column_exist('medication')):
            table_h.column_rename('name', 'medication')

        super().__register__(module)
        table_h = cls.__table_handler__(module)

    @classmethod
    def __setup__(cls):
        super(InpatientMedicationLog, cls).__setup__()

        # Do not cache default_key as it depends on time
        cls.__rpc__['default_get'].cache = None


class InpatientDiet (ModelSQL, ModelView):
    'Inpatient Diet'
    __name__ = "gnuhealth.inpatient.diet"

    registration = fields.Many2One(
        'gnuhealth.inpatient.registration',
        'Registration Code')
    diet = fields.Many2One('gnuhealth.diet.therapeutic', 'Diet', required=True)
    remarks = fields.Text(
        'Remarks / Directions',
        help='specific remarks for this diet / patient')

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name
        if (table_h.column_exist('name')
                and not table_h.column_exist('registration')):
            table_h.column_rename('name', 'registration')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class InpatientMeal (ModelSQL, ModelView):
    'Inpatient Meal'
    __name__ = "gnuhealth.inpatient.meal"

    meal = fields.Many2One(
        'product.product', 'Meal', required=True,
        help='Meal')

    diet_therapeutic = fields.Many2One(
        'gnuhealth.diet.therapeutic',
        'Diet')

    diet_belief = fields.Many2One('gnuhealth.diet.belief', 'Belief')

    diet_vegetarian = fields.Many2One(
        'gnuhealth.vegetarian_types',
        'Vegetarian')

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    @staticmethod
    def default_institution():
        return get_institution()

    def get_rec_name(self, name):
        if self.meal:
            return self.meal.name

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to meal
        if (table_h.column_exist('name')
                and not table_h.column_exist('meal')):
            table_h.column_rename('name', 'meal')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class InpatientMealOrderItem (ModelSQL, ModelView):
    'Inpatient Meal Item'
    __name__ = "gnuhealth.inpatient.meal.order.item"

    meal_order = fields.Many2One(
        'gnuhealth.inpatient.meal.order',
        'Meal Order')

    meal = fields.Many2One('gnuhealth.inpatient.meal', 'Meal')
    remarks = fields.Char('Remarks')

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to meal_order
        if (table_h.column_exist('name')
                and not table_h.column_exist('meal_order')):
            table_h.column_rename('name', 'meal_order')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class InpatientMealOrder (ModelSQL, ModelView):
    'Inpatient Meal Order'
    __name__ = "gnuhealth.inpatient.meal.order"

    registration = fields.Many2One(
        'gnuhealth.inpatient.registration',
        'Registration Code', domain=[('state', '=', 'hospitalized')],
        required=True)

    mealtime = fields.Selection((
        (None, ''),
        ('breakfast', 'Breakfast'),
        ('lunch', 'Lunch'),
        ('dinner', 'Dinner'),
        ('snack', 'Snack'),
        ('special', 'Special order'),
    ), 'Meal time', required=True, sort=False)

    meal_item = fields.One2Many(
        'gnuhealth.inpatient.meal.order.item', 'meal_order',
        'Items')

    meal_order = fields.Char('Order', readonly=True)

    health_professional = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Health Prof')

    remarks = fields.Text('Remarks')

    meal_warning = fields.Boolean(
        'Warning',
        help="The patient has special needs on meals")

    meal_warning_ack = fields.Boolean(
        'Ack',
        help="Check if you have verified the warnings on the"
        " patient meal items")

    order_date = fields.DateTime('Date', help='Order date', required=True)

    state = fields.Selection((
        (None, ''),
        ('draft', 'Draft'),
        ('cancelled', 'Cancelled'),
        ('ordered', 'Ordered'),
        ('processing', 'Processing'),
        ('done', 'Done'),
    ), 'Status', readonly=True)

    @staticmethod
    def default_order_date():
        return datetime.now()

    @staticmethod
    def default_state():
        return 'draft'

    @staticmethod
    def default_health_professional():
        return get_health_professional()

    @classmethod
    def generate_code(cls, **pattern):
        Config = Pool().get('gnuhealth.sequences')
        config = Config(1)
        sequence = config.get_multivalue(
            'inpatient_meal_order_sequence', **pattern)
        if sequence:
            return sequence.get()

    @classmethod
    def create(cls, vlist):
        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('meal_order'):
                values['meal_order'] = cls.generate_code()
        return super(InpatientMealOrder, cls).create(vlist)

    @classmethod
    def validate(cls, meal_orders):
        super(InpatientMealOrder, cls).validate(meal_orders)
        for meal_order in meal_orders:
            meal_order.check_meal_order_warning()
            meal_order.check_health_professional()

    def check_meal_order_warning(self):
        if not self.meal_warning_ack and self.meal_warning:
            raise SpecialMealNeeds(
                gettext('health_inpatient.msg_special_meal_needs'))

    def check_health_professional(self):
        if not self.health_professional:
            raise NoAssociatedHealthProfessional(
                gettext('health.msg_no_associated_health_professional'))

    @fields.depends('registration', '_parent_registration.patient')
    def on_change_registration(self):
        if self.registration:
            # Trigger the warning if the patient
            # has special needs on meals (religion / philosophy )
            if (self.registration.patient.vegetarian_type or
                    self.registration.patient.diet_belief):
                self.meal_warning = True
                self.meal_warning_ack = False

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to registration
        if (table_h.column_exist('name')
                and not table_h.column_exist('registration')):
            table_h.column_rename('name', 'registration')

        super().__register__(module)
        table_h = cls.__table_handler__(module)

    @classmethod
    def __setup__(cls):
        super(InpatientMealOrder, cls).__setup__()
        cls._buttons.update({
            'cancel': {'invisible': Not(Equal(Eval('state'), 'ordered'))}
        })

        cls._buttons.update({
            'generate': {
                'invisible': Or(Equal(Eval('state'), 'ordered'),
                                Equal(Eval('state'), 'done'))},
        })

        cls._buttons.update({
            'done': {'invisible': Not(Equal(Eval('state'), 'ordered'))}
        })

        t = cls.__table__()
        cls._sql_constraints = [
            ('meal_order_uniq', Unique(t, t.meal_order),
                'The Meal Order code already exists'),
        ]

        # Do not cache default_key as it depends on time
        cls.__rpc__['default_get'].cache = None

    @classmethod
    @ModelView.button
    def generate(cls, mealorders):
        cls.write(mealorders, {
            'state': 'ordered'})

    @classmethod
    @ModelView.button
    def cancel(cls, mealorders):
        cls.write(mealorders, {
            'state': 'cancelled'})

    @classmethod
    @ModelView.button
    def done(cls, mealorders):
        cls.write(mealorders, {
            'state': 'done'})


# Class : PatientRounding
# Assess the patient and evironment periodically
# Usually done by nurses

class PatientRounding(ModelSQL, ModelView):
    'Patient Rounding'
    __name__ = 'gnuhealth.patient.rounding'

    STATES = {'readonly': Eval('state') == 'done'}

    registration = fields.Many2One(
        'gnuhealth.inpatient.registration',
        'Registration Code', required=True, states=STATES)
    code = fields.Char('Code', readonly=True)
    health_professional = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Health Prof', readonly=True)
    evaluation_start = fields.DateTime('Start', required=True, states=STATES)
    evaluation_end = fields.DateTime('End', readonly=True)

    state = fields.Selection([
        (None, ''),
        ('draft', 'In Progress'),
        ('done', 'Done'),
    ], 'State', readonly=True)

    environmental_assessment = fields.Char(
        'Environment', help="Environment"
        " assessment . State any disorder in the room.", states=STATES)

    weight = fields.Integer(
        'Weight',
        help="Measured weight, in kg", states=STATES)

    # The 6 P's of rounding
    pain = fields.Boolean(
        'Pain',
        help="Check if the patient is in pain", states=STATES)

    pain_level = fields.Integer(
        'Pain level',
        help="Enter the pain level, from 1 to 10.",
        states={'readonly': Eval('state') == 'done'})

    # Use by round_report template
    def get_report_pain_and_level(self):
        if self.pain and self.pain_level:
            return gettext('health_nursing.msg_report_pain_level',
                           pain_level=str(self.pain_level))
        elif self.pain:
            return gettext('health_nursing.msg_report_pain_yes')
        else:
            return gettext('health_nursing.msg_report_pain_no')

    potty = fields.Boolean(
        'Potty', help="Check if the patient needs to "
        "urinate / defecate", states=STATES)
    position = fields.Boolean(
        'Position', help="Check if the patient needs to "
        "be repositioned or is unconfortable", states=STATES)
    proximity = fields.Boolean(
        'Proximity', help="Check if personal items, "
        "water, alarm, ... are not in easy reach", states=STATES)
    pump = fields.Boolean(
        'Pumps', help="Check if there is any issues with "
        "the pumps - IVs ...", states=STATES)
    personal_needs = fields.Boolean(
        'Personal needs', help="Check if the "
        "patient requests anything", states=STATES)

    # Vital Signs
    systolic = fields.Integer('Systolic Pressure', states=STATES)
    diastolic = fields.Integer('Diastolic Pressure', states=STATES)
    bpm = fields.Integer(
        'Heart Rate',
        help='Heart rate expressed in beats per minute', states=STATES)
    respiratory_rate = fields.Integer(
        'Respiratory Rate',
        help='Respiratory rate expressed in breaths per minute', states=STATES)
    osat = fields.Integer(
        'Oxygen Saturation',
        help='Oxygen Saturation(arterial).', states=STATES)
    temperature = fields.Float(
        'Temperature',
        help='Temperature in celsius', states=STATES)

    # Diuresis

    diuresis = fields.Integer('Diuresis', help="volume in ml", states=STATES)
    urinary_catheter = fields.Boolean('Urinary Catheter', states=STATES)

    # Glycemia
    glycemia = fields.Integer(
        'Glycemia', help='Blood Glucose level', states=STATES)

    depression = fields.Boolean(
        'Depression signs', help="Check this if the "
        "patient shows signs of depression", states=STATES)
    evolution = fields.Selection(
        [
            (None, ''),
            ('n', 'Status Quo'),
            ('i', 'Improving'),
            ('w', 'Worsening'),
        ], 'Evolution', help="Check your judgement of current "
        "patient condition", sort=False, states=STATES)

    evolution_str = evolution.translated('evolution')

    round_summary = fields.Text('Round Summary', states=STATES)

    signed_by = fields.Many2One(
        'gnuhealth.healthprofessional', 'Signed by', readonly=True,
        states={'invisible': Equal(Eval('state'), 'draft')},
        help="Health Professional that signed the rounding")

    warning = fields.Boolean(
        'Warning', help="Check this box to alert the "
        "supervisor about this patient rounding. A warning icon will be shown "
        "in the rounding list", states=STATES)
    warning_icon = fields.Function(
        fields.Char('Warning Icon'), 'get_warn_icon')

    rounding_procedures = fields.One2Many(
        'gnuhealth.patient.procedure', 'reference', 'Procedures',
        domain=[
            ('patient', '=', Eval('patient')),
            ('ctx', '=', 'rounding'),
            # ('pdate', '=', Eval('evaluation_start')),
        ],
        depends=['registration', 'patient'],
        help='Procedures done during the rounding')

    # Deprecated in 5.0 by rounding_procedures
    procedures = fields.One2Many(
        'gnuhealth.rounding_procedure', 'rounding',
        'Procedures', help="List of the procedures in this rounding. Please "
        "enter the first one as the main procedure", states=STATES)

    report_start_date = fields.Function(
        fields.Date('Start Date'),
        'get_report_start_date')
    report_start_time = fields.Function(
        fields.Time('Start Time'),
        'get_report_start_time')
    report_end_date = fields.Function(
        fields.Date('End Date'),
        'get_report_end_date')
    report_end_time = fields.Function(
        fields.Time('End Time'),
        'get_report_end_time')

    patient = fields.Function(fields.Many2One(
        'gnuhealth.patient', 'Patient'), 'get_patient')

    def get_patient(self, name):
        if (self.registration):
            return self.registration.patient

    # Show patient upon entering the registration
    @fields.depends('registration')
    def on_change_registration(self):
        if (self.registration):
            self.patient = self.registration.patient

    @staticmethod
    def default_health_professional():
        return get_health_professional()

    @staticmethod
    def default_evaluation_start():
        return datetime.now()

    @staticmethod
    def default_state():
        return 'draft'

    @classmethod
    @ModelView.button
    def end_rounding(cls, roundings):
        # End the rounding

        # Change the state of the rounding to "Done"
        signing_hp = get_health_professional()

        cls.write(roundings, {
            'state': 'done',
            'signed_by': signing_hp,
            'evaluation_end': datetime.now()
        })

    @classmethod
    def generate_code(cls, **pattern):
        Config = Pool().get('gnuhealth.sequences')
        config = Config(1)
        sequence = config.get_multivalue(
            'patient_rounding_sequence', **pattern)
        if sequence:
            return sequence.get()

    def get_rec_name(self, name):
        if self.code:
            return f"{self.code}"

    @classmethod
    def create(cls, vlist):
        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('code'):
                values['code'] = cls.generate_code()
        return super(PatientRounding, cls).create(vlist)

    @classmethod
    def validate(cls, roundings):
        super(PatientRounding, cls).validate(roundings)
        for rounding in roundings:
            rounding.check_health_professional()

    def check_health_professional(self):
        if not self.health_professional:
            raise NoAssociatedHealthProfessional(
                gettext('health.msg_no_associated_health_professional'))

    def get_report_start_date(self, name):
        Company = Pool().get('company.company')

        timezone = None
        company_id = Transaction().context.get('company')
        if company_id:
            company = Company(company_id)
            if company.timezone:
                timezone = pytz.timezone(company.timezone)

        dt = self.evaluation_start
        return datetime.astimezone(
            dt.replace(tzinfo=pytz.utc), timezone).date()

    def get_report_start_time(self, name):
        Company = Pool().get('company.company')

        timezone = None
        company_id = Transaction().context.get('company')
        if company_id:
            company = Company(company_id)
            if company.timezone:
                timezone = pytz.timezone(company.timezone)

        dt = self.evaluation_start
        return datetime.astimezone(
            dt.replace(tzinfo=pytz.utc), timezone).time()

    def get_report_end_date(self, name):
        Company = Pool().get('company.company')

        timezone = None
        company_id = Transaction().context.get('company')
        if company_id:
            company = Company(company_id)
            if company.timezone:
                timezone = pytz.timezone(company.timezone)

        dt = self.evaluation_end
        return datetime.astimezone(
            dt.replace(tzinfo=pytz.utc), timezone).date()

    def get_report_end_time(self, name):
        Company = Pool().get('company.company')

        timezone = None
        company_id = Transaction().context.get('company')
        if company_id:
            company = Company(company_id)
            if company.timezone:
                timezone = pytz.timezone(company.timezone)

        dt = self.evaluation_end
        return datetime.astimezone(
            dt.replace(tzinfo=pytz.utc), timezone).time()

    def get_warn_icon(self, name):
        if self.warning:
            return 'gnuhealth-warning'

    @classmethod
    def __setup__(cls):
        super(PatientRounding, cls).__setup__()
        cls._buttons.update({
            'end_rounding': {
                'invisible': ~Eval('state').in_(['draft']),
            }})

        cls._order.insert(0, ('evaluation_start', 'DESC'))

        # Do not cache default_key as it depends on time
        cls.__rpc__['default_get'].cache = None

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to registration
        if (table_h.column_exist('name')
                and not table_h.column_exist('registration')):
            table_h.column_rename('name', 'registration')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


# Deprecated in GH 5.0 by general health.PatientProcedure
class RoundingProcedure(ModelSQL, ModelView):
    'Rounding - Procedure'
    __name__ = 'gnuhealth.rounding_procedure'

    rounding = fields.Many2One('gnuhealth.patient.rounding', 'Rounding')
    procedure = fields.Many2One(
        'gnuhealth.procedure', 'Code', required=True,
        help="Procedure Code, for example ICD-10-PCS Code 7-character string")
    notes = fields.Text('Notes')

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to rounding
        if (table_h.column_exist('name')
                and not table_h.column_exist('rounding')):
            table_h.column_rename('name', 'rounding')

        super().__register__(module)
        table_h = cls.__table_handler__(module)



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                           HEALTH package                              #
#              wizard_update_patient_disease_info.py: wizard            #
#########################################################################
from trytond.wizard import Wizard, StateView, Button, StateTransition
from trytond.transaction import Transaction
from trytond.pool import Pool

from trytond.modules.health.core import (parse_compute_age)
from trytond.i18n import gettext

from ..exceptions import (PatientDiseaseAlreadyExists)

__all__ = ['UpdatePatientDiseaseInfo']


class UpdatePatientDiseaseInfo(Wizard):
    __name__ = 'gnuhealth.update_patient_disease_info'

    start = StateTransition()

    update_disease = StateView(
        'gnuhealth.patient.disease',
        'health.gnuhealth_patient_diseases_view_form_for_wizard', [
            Button('Cancel', 'end', 'tryton-cancel'),
            Button('Save', 'save', 'tryton-ok', default=True)])

    save = StateTransition()

    def transition_start(self):
        pool = Pool()
        Disease = Pool().get('gnuhealth.patient.disease')
        Evaluation = pool.get('gnuhealth.patient.evaluation')

        evaluation = Evaluation.browse(
            [Transaction().context.get('active_id')])[0]

        existing_disease = Disease.search(
            [('patient', '=', evaluation.patient),
             ('pathology', '=', evaluation.diagnosis),
             ('diagnosed_date', '=', evaluation.evaluation_endtime.date())])

        if existing_disease:
            raise PatientDiseaseAlreadyExists(
                gettext('health.msg_patient_disease_already_exists')
            )
            return 'end'
        else:
            return 'update_disease'

    def default_update_disease(self, fields):
        pool = Pool()
        Evaluation = pool.get('gnuhealth.patient.evaluation')

        evaluation = Evaluation.browse(
            [Transaction().context.get('active_id')])[0]

        return {'patient': (evaluation.patient
                            and evaluation.patient.id),
                'age': ((evaluation.patient and evaluation.patient.age)
                        and parse_compute_age(evaluation.patient.age)[0]),
                'age_str': (evaluation.patient
                            and evaluation.patient.age),
                'pathology': (evaluation.diagnosis
                              and evaluation.diagnosis.id),
                'institution': (evaluation.institution
                                and evaluation.institution.id),
                'diagnosed_date': evaluation.evaluation_endtime}

    def transition_save(self):
        Disease = Pool().get('gnuhealth.patient.disease')
        Disease.save([self.update_disease])
        return 'end'



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2011-2012 Sebastian Marro <smarro@thymbra.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                  HEALTH INPATIENT CALENDAR PACKAGE                    #
#             health_inpatient_calendar.py: main module                 #
#########################################################################

from trytond.model import fields
from trytond.pool import Pool, PoolMeta


__all__ = ['HospitalBed', 'InpatientRegistration']


class HospitalBed(metaclass=PoolMeta):
    __name__ = "gnuhealth.hospital.bed"

    calendar = fields.Many2One(
        'calendar.calendar', 'Calendar',
        help="A calendar can be associated to a bed. To use this "
        "functionality, it needs to be created via health -> calendars")


class InpatientRegistration(metaclass=PoolMeta):
    __name__ = 'gnuhealth.inpatient.registration'

    event = fields.Many2One(
        'calendar.event', 'Calendar Event', readonly=True,
        help="Calendar Event associated to this hospitalization")

    @classmethod
    def confirmed(cls, registrations):
        super(InpatientRegistration, cls).confirmed(registrations)

        Event = Pool().get('calendar.event')

        for inpatient_registration in registrations:
            if inpatient_registration.bed.calendar:
                if not inpatient_registration.event:
                    bed = inpatient_registration.bed.product.code + ": "
                    events = Event.create([{
                        'dtstart': inpatient_registration.hospitalization_date,
                        'dtend': inpatient_registration.discharge_date,
                        'calendar': inpatient_registration.bed.calendar.id,
                        'summary':
                            bed + inpatient_registration.patient.party.rec_name
                    }])
                    cls.write(
                        [inpatient_registration],
                        {'event': events[0].id})

    @classmethod
    def discharge(cls, registrations):
        super(InpatientRegistration, cls).discharge(registrations)

        Event = Pool().get('calendar.event')

        for inpatient_registration in registrations:
            if inpatient_registration.event:
                Event.delete([inpatient_registration.event])

    @classmethod
    def write(cls, registrations, values):
        Event = Pool().get('calendar.event')
        Patient = Pool().get('gnuhealth.patient')
        HospitalBed = Pool().get('gnuhealth.hospital.bed')

        for inpatient_registration in registrations:
            if inpatient_registration.event:
                if 'hospitalization_date' in values:
                    Event.write([inpatient_registration.event], {
                        'dtstart': values['hospitalization_date'],
                    })
                if 'discharge_date' in values:
                    Event.write([inpatient_registration.event], {
                        'dtend': values['discharge_date'],
                    })
                if 'bed' in values:
                    bed = HospitalBed(values['bed'])
                    Event.write([inpatient_registration.event], {
                        'calendar': bed.calendar.id,
                    })
                if 'patient' in values:
                    patient = Patient(values['patient'])
                    bed = inpatient_registration.bed.product.code + ": "
                    Event.write([inpatient_registration.event], {
                        'summary': bed + patient.party.rec_name,
                    })

        return super(InpatientRegistration, cls).write(registrations, values)

    @classmethod
    def delete(cls, registrations):
        Event = Pool().get('calendar.event')

        for inpatient_registration in registrations:
            if inpatient_registration.event:
                Event.delete([inpatient_registration.event])
        return super(InpatientRegistration, cls).delete(registrations)



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                        HEALTH PEDIATRICS package                      #
#                   health_pediatrics.py: main module                   #
#########################################################################

from trytond.model import ModelView, ModelSQL, fields, Unique
from trytond.transaction import Transaction
from trytond.pool import Pool, PoolMeta
from datetime import datetime
from trytond.pyson import Eval, Not, Equal
from trytond.modules.health.core import get_health_professional


__all__ = ['Newborn', 'NeonatalApgar', 'NeonatalMedication',
           'NeonatalCongenitalDiseases', 'PediatricSymptomsChecklist']


class Newborn(ModelSQL, ModelView):
    'Newborn Information'
    __name__ = 'gnuhealth.newborn'

    STATES = {'readonly': Eval('state') == 'signed'}

    # We no longer need the legacy newborn ID, since patient.puid is
    # used. for backward compatibility reasons, we keep it.
    name = fields.Char('Newborn ID', states=STATES)
    patient = fields.Many2One(
        'gnuhealth.patient', 'Baby', required=True, states=STATES,
        help="Patient associated to this newborn")

    mother = fields.Many2One('gnuhealth.patient', 'Mother', states=STATES)
    newborn_name = fields.Char('Name at Birth', states=STATES)
    birth_date = fields.DateTime(
        'DoB', required=True,
        help="Date and Time of birth", states=STATES)
    photo = fields.Binary('Picture', states=STATES)

    # Sex / Gender at birth.
    sex = fields.Selection(
        [
            ('m', 'Male'),
            ('f', 'Female'),
        ], 'Sex', sort=False, required=True,
        help="Sex at birth. It might differ from the current patient"
        " gender. This is the biological sex.", states=STATES)

    sex_str = sex.translated('sex')

    state = fields.Selection([
        (None, ''),
        ('draft', 'draft'),
        ('signed', 'Signed'),
    ], 'State', readonly=True, sort=False)

    cephalic_perimeter = fields.Integer(
        'CP',
        help="Cephalic Perimeter in centimeters (cm)", states=STATES)
    length = fields.Integer(
        'Length',
        help="Length in centimeters (cm)", states=STATES)
    weight = fields.Integer(
        'Weight',
        help="Weight in grams (g)", states=STATES)

    # Used by newborn_card report template, the page of newborn_card
    # report is very small, so it requires compression of information.
    def get_report_length_and_weight(self):
        if self.length and self.weight:
            # I think 'cm' and 'g' do not need to translate.
            return f'{self.length}cm, {self.weight}g'
        elif self.length:
            return f'{self.length}cm'
        elif self.weight:
            return f'{self.weight}g'
        else:
            return ''

    apgar1 = fields.Integer('APGAR 1st minute', states=STATES)
    apgar5 = fields.Integer('APGAR 5th minute', states=STATES)
    apgar_scores = fields.One2Many(
        'gnuhealth.neonatal.apgar', 'newborn',
        'APGAR scores', states=STATES)
    meconium = fields.Boolean('Meconium', states=STATES)

    # Deprecated. Use Patient conditions directly
    congenital_diseases = fields.One2Many(
        'gnuhealth.patient.disease',
        'newborn_id', 'Congenital diseases', states=STATES)

    reanimation_stimulation = fields.Boolean('Stimulation', states=STATES)
    reanimation_aspiration = fields.Boolean('Aspiration', states=STATES)
    reanimation_intubation = fields.Boolean('Intubation', states=STATES)
    reanimation_mask = fields.Boolean('Mask', states=STATES)
    reanimation_oxygen = fields.Boolean('Oxygen', states=STATES)
    test_vdrl = fields.Boolean('VDRL', states=STATES)
    test_toxo = fields.Boolean('Toxoplasmosis', states=STATES)
    test_chagas = fields.Boolean('Chagas', states=STATES)
    test_billirubin = fields.Boolean('Billirubin', states=STATES)
    test_audition = fields.Boolean('Audition', states=STATES)
    test_metabolic = fields.Boolean(
        'Metabolic ("heel stick screening")',
        help="Test for Fenilketonuria, Congenital Hypothyroidism, "
        "Quistic Fibrosis, Galactosemia", states=STATES)
    neonatal_ortolani = fields.Boolean('Positive Ortolani', states=STATES)
    neonatal_barlow = fields.Boolean('Positive Barlow', states=STATES)
    neonatal_hernia = fields.Boolean('Hernia', states=STATES)
    neonatal_ambiguous_genitalia = fields.Boolean(
        'Ambiguous Genitalia',
        states=STATES)
    neonatal_erbs_palsy = fields.Boolean('Erbs Palsy', states=STATES)
    neonatal_hematoma = fields.Boolean('Hematomas', states=STATES)
    neonatal_talipes_equinovarus = fields.Boolean(
        'Talipes Equinovarus',
        states=STATES)
    neonatal_polydactyly = fields.Boolean('Polydactyly', states=STATES)
    neonatal_syndactyly = fields.Boolean('Syndactyly', states=STATES)
    neonatal_moro_reflex = fields.Boolean('Moro Reflex', states=STATES)
    neonatal_grasp_reflex = fields.Boolean('Grasp Reflex', states=STATES)
    neonatal_stepping_reflex = fields.Boolean(
        'Stepping Reflex',
        states=STATES)
    neonatal_babinski_reflex = fields.Boolean(
        'Babinski Reflex',
        states=STATES)
    neonatal_blink_reflex = fields.Boolean('Blink Reflex', states=STATES)
    neonatal_sucking_reflex = fields.Boolean('Sucking Reflex', states=STATES)
    neonatal_swimming_reflex = fields.Boolean(
        'Swimming Reflex',
        states=STATES)
    neonatal_tonic_neck_reflex = fields.Boolean(
        'Tonic Neck Reflex',
        states=STATES)
    neonatal_rooting_reflex = fields.Boolean(
        'Rooting Reflex',
        states=STATES)
    neonatal_palmar_crease = fields.Boolean(
        'Transversal Palmar Crease',
        states=STATES)

    # Deprecated. Use Patient medication directly
    medication = fields.One2Many(
        'gnuhealth.patient.medication',
        'newborn_id', 'Medication')

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Health Prof',
        help="Health professional", readonly=True)

    signed_by = fields.Many2One(
        'gnuhealth.healthprofessional', 'Signed by', readonly=True,
        states={
            'invisible': Not(Equal(Eval('state'), 'signed'))
        },
        help="Health Professional that signed this document")

    dismissed = fields.DateTime('Discharged', states=STATES)
    notes = fields.Text('Notes', states=STATES)

    # Deprecated. the following fields will be removed in 2.8
    # Decease information on fetus / newborn are linked now in
    # the obstetrics evaluation (prenatal) or patient if result of pregnancy
    # was a live birth.
    # The information is no longer shown at the view.
    # Fields to be removed : bd, died_at_delivery, died_at_the_hospital
    # died_being_transferred, tod, cod

    bd = fields.Boolean('Stillbirth')
    died_at_delivery = fields.Boolean('Died at delivery room')
    died_at_the_hospital = fields.Boolean('Died at the hospital')
    died_being_transferred = fields.Boolean(
        'Died being transferred',
        help="The baby died being transferred to another health institution")
    tod = fields.DateTime('Time of Death')
    cod = fields.Many2One('gnuhealth.pathology', 'Cause of death')

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    @staticmethod
    def default_state():
        return 'draft'

    @classmethod
    def __setup__(cls):
        super(Newborn, cls).__setup__()

        t = cls.__table__()
        cls._sql_constraints = [
            ('name_uniq', Unique(t, t.name),
             'The Newborn ID must be unique'),
            ('patient_uniq', Unique(t, t.patient),
             'There is already a newborn record for this patient'),
        ]
        cls._order.insert(0, ('birth_date', 'DESC'))

        cls._buttons.update({
            'sign_newborn': {'invisible': Equal(Eval('state'), 'signed')}
        })

    @classmethod
    @ModelView.button
    def sign_newborn(cls, newborns):
        signing_hp = get_health_professional()

        cls.write(newborns, {
            'state': 'signed',
            'signed_by': signing_hp,
        })

    @classmethod
    def write(cls, newborns, values):
        pool = Pool()
        Patient = pool.get('gnuhealth.patient')
        Party = pool.get('party.party')

        party = []
        patient = []

        for newborn in newborns:

            newborn_patient_id = newborn.patient.id

            person = Patient.browse([newborn_patient_id])[0].party
            pat = Patient.browse([newborn_patient_id])[0]

            # Update the birth date on the party model upon WRITING it on the
            # newborn model

            born_date = datetime.date(newborn.birth_date)

            party.append(person)

            Party.write(party, {'dob': born_date})

            # Update the biological sex on the patient model upon WRITING
            # it on the newborn model

            if values.get('sex'):
                biological_sex = values.get('sex')

                patient.append(pat)

                Patient.write(patient, {'biological_sex': biological_sex})

        return super(Newborn, cls).write(newborns, values)

    @classmethod
    def create(cls, vlist):
        pool = Pool()
        vlist = [x.copy() for x in vlist]
        Patient = pool.get('gnuhealth.patient')
        Party = pool.get('party.party')

        party = []
        patient = []

        for values in vlist:
            newborn_patient_id = values['patient']

            person = Patient.browse([newborn_patient_id])[0].party
            pat = Patient.browse([newborn_patient_id])[0]

            # Update the birth date on the party model upon CREATING it on the
            # newborn model

            born_date = datetime.date(values['birth_date'])

            party.append(person)

            Party.write(party, {'dob': born_date})

            # Update the biological sex on the patient model upon CREATING
            # it on the newborn model

            if values.get('sex'):
                biological_sex = values.get('sex')

                patient.append(pat)

                Patient.write(patient, {'biological_sex': biological_sex})

        return super(Newborn, cls).create(vlist)

    @classmethod
    def search_rec_name(cls, name, clause):
        if clause[1].startswith('!') or clause[1].startswith('not '):
            bool_op = 'AND'
        else:
            bool_op = 'OR'
        return [bool_op, ('name',) + tuple(clause[1:]),
                ('patient',) + tuple(clause[1:]),
                ]


class NeonatalApgar(ModelSQL, ModelView):
    'Neonatal APGAR Score'
    __name__ = 'gnuhealth.neonatal.apgar'

    newborn = fields.Many2One('gnuhealth.newborn', 'Newborn')

    apgar_minute = fields.Integer('Minute', required=True)

    apgar_appearance = fields.Selection([
        ('0', 'central cyanosis'),
        ('1', 'acrocyanosis'),
        ('2', 'no cyanosis'),
    ], 'Appearance', required=True)

    apgar_pulse = fields.Selection([
        ('0', 'Absent'),
        ('1', '< 100'),
        ('2', '> 100'),
    ], 'Pulse', required=True)

    apgar_grimace = fields.Selection([
        ('0', 'No response to stimulation'),
        ('1', 'grimace when stimulated'),
        ('2', 'cry or pull away when stimulated'),
    ], 'Grimace', required=True, sort=False)

    apgar_activity = fields.Selection([
        ('0', 'None'),
        ('1', 'Some flexion'),
        ('2', 'flexed arms and legs'),
    ], 'Activity', required=True, sort=False)

    apgar_respiration = fields.Selection([
        ('0', 'Absent'),
        ('1', 'Weak / Irregular'),
        ('2', 'strong'),
    ], 'Respiration', required=True, sort=False)

    apgar_score = fields.Integer('APGAR Score')

    @fields.depends(
        'apgar_respiration', 'apgar_activity', 'apgar_grimace',
        'apgar_pulse', 'apgar_appearance')
    def on_change_with_apgar_score(self):
        apgar_appearance = self.apgar_appearance or '0'
        apgar_pulse = self.apgar_pulse or '0'
        apgar_grimace = self.apgar_grimace or '0'
        apgar_activity = self.apgar_activity or '0'
        apgar_respiration = self.apgar_respiration or '0'

        apgar_score = int(apgar_appearance) + int(apgar_pulse) + \
            int(apgar_grimace) + int(apgar_activity) + int(apgar_respiration)

        return apgar_score

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to newborn
        if (table_h.column_exist('name')
                and not table_h.column_exist('newborn')):
            table_h.column_rename('name', 'newborn')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


# Deprecated in 3.0 - Use the main patient form
class NeonatalMedication(metaclass=PoolMeta):
    __name__ = 'gnuhealth.patient.medication'

    newborn_id = fields.Many2One('gnuhealth.newborn', 'Newborn')


# Deprecated in 3.0  - Use main patient form
class NeonatalCongenitalDiseases(metaclass=PoolMeta):
    __name__ = 'gnuhealth.patient.disease'

    newborn_id = fields.Many2One('gnuhealth.newborn', 'Newborn')


class PediatricSymptomsChecklist(ModelSQL, ModelView):
    'Pediatric Symptoms Checklist'
    __name__ = 'gnuhealth.patient.psc'

    patient = fields.Many2One('gnuhealth.patient', 'Patient', required=True)

    evaluation_date = fields.Many2One(
        'gnuhealth.appointment', 'Appointment',
        help="Enter or select the date / ID of the appointment related to "
        "this evaluation")

    evaluation_start = fields.DateTime('Date', required=True)

    user_id = fields.Many2One('res.user', 'Health Prof', readonly=True)

    notes = fields.Text('Notes')

    psc_aches_pains = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Complains of aches and pains', sort=False)

    psc_spend_time_alone = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Spends more time alone', sort=False)

    psc_tires_easily = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Tires easily, has little energy', sort=False)

    psc_fidgety = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Fidgety, unable to sit still', sort=False)

    psc_trouble_with_teacher = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Has trouble with teacher', sort=False)

    psc_less_interest_in_school = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Less interested in school', sort=False)

    psc_acts_as_driven_by_motor = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Acts as if driven by a motor', sort=False)

    psc_daydreams_too_much = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Daydreams too much', sort=False)

    psc_distracted_easily = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Distracted easily', sort=False)

    psc_afraid_of_new_situations = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Is afraid of new situations', sort=False)

    psc_sad_unhappy = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Feels sad, unhappy', sort=False)

    psc_irritable_angry = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Is irritable, angry', sort=False)

    psc_feels_hopeless = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Feels hopeless', sort=False)

    psc_trouble_concentrating = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Has trouble concentrating', sort=False)

    psc_less_interested_in_friends = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Less interested in friends', sort=False)

    psc_fights_with_others = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Fights with other children', sort=False)

    psc_absent_from_school = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Absent from school', sort=False)

    psc_school_grades_dropping = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'School grades dropping', sort=False)

    psc_down_on_self = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Is down on him or herself', sort=False)

    psc_visit_doctor_finds_ok = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Visits the doctor with doctor finding nothing wrong', sort=False)

    psc_trouble_sleeping = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Has trouble sleeping', sort=False)

    psc_worries_a_lot = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Worries a lot', sort=False)

    psc_wants_to_be_with_parents = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Wants to be with you more than before', sort=False)

    psc_feels_is_bad_child = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Feels he or she is bad', sort=False)

    psc_takes_unnecesary_risks = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Takes unnecessary risks', sort=False)

    psc_gets_hurt_often = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Gets hurt frequently', sort=False)

    psc_having_less_fun = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Seems to be having less fun', sort=False)

    psc_act_as_younger = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Acts younger than children his or her age', sort=False)

    psc_does_not_listen_to_rules = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Does not listen to rules', sort=False)

    psc_does_not_show_feelings = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Does not show feelings', sort=False)

    psc_does_not_get_people_feelings = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Does not get people feelings', sort=False)

    psc_teases_others = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Teases others', sort=False)

    psc_blames_others = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Blames others for his or her troubles', sort=False)

    psc_takes_things_from_others = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Takes things that do not belong to him or her', sort=False)

    psc_refuses_to_share = fields.Selection([
        (None, ''),
        ('0', 'Never'),
        ('1', 'Sometimes'),
        ('2', 'Often'),
    ], 'Refuses to share', sort=False)

    psc_total = fields.Integer('PSC Total')

    @staticmethod
    def default_user_id():
        User = Pool().get('res.user')
        user = User(Transaction().user)
        return int(user.id)

    @staticmethod
    def default_psc_total():
        return 0

    @fields.depends(
        'psc_aches_pains', 'psc_spend_time_alone',
        'psc_tires_easily', 'psc_fidgety', 'psc_trouble_with_teacher',
        'psc_less_interest_in_school', 'psc_acts_as_driven_by_motor',
        'psc_daydreams_too_much', 'psc_distracted_easily',
        'psc_afraid_of_new_situations', 'psc_sad_unhappy',
        'psc_irritable_angry', 'psc_feels_hopeless',
        'psc_trouble_concentrating', 'psc_less_interested_in_friends',
        'psc_fights_with_others', 'psc_absent_from_school',
        'psc_school_grades_dropping', 'psc_down_on_self',
        'psc_visit_doctor_finds_ok', 'psc_trouble_sleeping',
        'psc_worries_a_lot', 'psc_wants_to_be_with_parents',
        'psc_feels_is_bad_child', 'psc_takes_unnecesary_risks',
        'psc_gets_hurt_often', 'psc_having_less_fun',
        'psc_act_as_younger', 'psc_does_not_listen_to_rules',
        'psc_does_not_show_feelings',
        'psc_does_not_get_people_feelings',
        'psc_teases_others', 'psc_takes_things_from_others',
        'psc_refuses_to_share')
    def on_change_with_psc_total(self):

        psc_aches_pains = self.psc_aches_pains or '0'
        psc_spend_time_alone = self.psc_spend_time_alone or '0'
        psc_tires_easily = self.psc_tires_easily or '0'
        psc_fidgety = self.psc_fidgety or '0'
        psc_trouble_with_teacher = self.psc_trouble_with_teacher or '0'
        psc_less_interest_in_school = self.psc_less_interest_in_school or '0'
        psc_acts_as_driven_by_motor = self.psc_acts_as_driven_by_motor or '0'
        psc_daydreams_too_much = self.psc_daydreams_too_much or '0'
        psc_distracted_easily = self.psc_distracted_easily or '0'
        psc_afraid_of_new_situations = self.psc_afraid_of_new_situations or '0'
        psc_sad_unhappy = self.psc_sad_unhappy or '0'
        psc_irritable_angry = self.psc_irritable_angry or '0'
        psc_feels_hopeless = self.psc_feels_hopeless or '0'
        psc_trouble_concentrating = self.psc_trouble_concentrating or '0'
        psc_less_interested_in_friends = \
            self.psc_less_interested_in_friends or '0'
        psc_fights_with_others = self.psc_fights_with_others or '0'
        psc_absent_from_school = self.psc_absent_from_school or '0'
        psc_school_grades_dropping = self.psc_school_grades_dropping or '0'
        psc_down_on_self = self.psc_down_on_self or '0'
        psc_visit_doctor_finds_ok = self.psc_visit_doctor_finds_ok or '0'
        psc_trouble_sleeping = self.psc_trouble_sleeping or '0'
        psc_worries_a_lot = self.psc_worries_a_lot or '0'
        psc_wants_to_be_with_parents = self.psc_wants_to_be_with_parents or '0'
        psc_feels_is_bad_child = self.psc_feels_is_bad_child or '0'
        psc_takes_unnecesary_risks = self.psc_takes_unnecesary_risks or '0'
        psc_gets_hurt_often = self.psc_gets_hurt_often or '0'
        psc_having_less_fun = self.psc_having_less_fun or '0'
        psc_act_as_younger = self.psc_act_as_younger or '0'
        psc_does_not_listen_to_rules = self.psc_does_not_listen_to_rules or '0'
        psc_does_not_show_feelings = self.psc_does_not_show_feelings or '0'
        psc_does_not_get_people_feelings = \
            self.psc_does_not_get_people_feelings or '0'
        psc_teases_others = self.psc_teases_others or '0'
        psc_takes_things_from_others = self.psc_takes_things_from_others or '0'
        psc_refuses_to_share = self.psc_refuses_to_share or '0'

        psc_total = int(psc_aches_pains) + int(psc_spend_time_alone) + \
            int(psc_tires_easily) + int(psc_fidgety) + \
            int(psc_trouble_with_teacher) + \
            int(psc_less_interest_in_school) + \
            int(psc_acts_as_driven_by_motor) + \
            int(psc_daydreams_too_much) + int(psc_distracted_easily) + \
            int(psc_afraid_of_new_situations) + int(psc_sad_unhappy) + \
            int(psc_irritable_angry) + int(psc_feels_hopeless) + \
            int(psc_trouble_concentrating) + \
            int(psc_less_interested_in_friends) + \
            int(psc_fights_with_others) + int(psc_absent_from_school) + \
            int(psc_school_grades_dropping) + int(psc_down_on_self) + \
            int(psc_visit_doctor_finds_ok) + int(psc_trouble_sleeping) + \
            int(psc_worries_a_lot) + int(psc_wants_to_be_with_parents) + \
            int(psc_feels_is_bad_child) + int(psc_takes_unnecesary_risks) + \
            int(psc_gets_hurt_often) + int(psc_having_less_fun) + \
            int(psc_act_as_younger) + int(psc_does_not_listen_to_rules) + \
            int(psc_does_not_show_feelings) + \
            int(psc_does_not_get_people_feelings) + \
            int(psc_teases_others) + \
            int(psc_takes_things_from_others) + \
            int(psc_refuses_to_share)

        return psc_total



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .test_health_pediatrics_growth_charts import suite


__all__ = ['suite']



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                   HEALTH CONTACT TRACING package                      #
#             health_contact_tracing.py: main module                    #
#########################################################################

from trytond.model import ModelView, ModelSQL, fields


__all__ = ['ContactTracing']


class ContactTracing(ModelSQL, ModelView):
    'Person Contacts tracing'
    __name__ = 'gnuhealth.contact_tracing'

    patient = fields.Many2One(
        'gnuhealth.patient', 'Source', required=True,
        help="Person suspected or confirmed")

    pathology = fields.Many2One(
        'gnuhealth.pathology', 'Disease',
        help='Disease to trace')

    contact = fields.Many2One(
        'party.party', 'Contact',
        domain=[('is_person', '=', True)],
        help="Person that the patient has contacted")

    contact_date = fields.DateTime("Date")

    du = fields.Many2One(
        'gnuhealth.du', 'Dom.Unit',
        help="Domiciliary Unit")

    operational_sector = fields.Many2One(
        'gnuhealth.operational_sector', 'Op.Sector',
        help="Operational / Sanitary region")

    exposure_risk = fields.Selection((
        ('low', 'Low'),
        ('high', 'High'),
        ('na', 'Not available'),
    ), 'Exposure', required=True, sort=False)

    exposure_time = fields.Integer(
        "Exposure time",
        help="Exposure time in minutes")

    status = fields.Selection((
        ('unreached', 'Unreached'),
        ('followingup', 'Following up'),
        ('na', 'Not available'),
    ), 'Status', required=True, sort=False,
        help="Unreached: The contact has not been reached yet."
        "\nFollowing up: The contact has been traced, demographics information"
        " has been created and followup evaluations status are stored in the"
        " evaluations")

    context = fields.Char(
        "Context",
        help="Airport, meeting, concert, ...")

    comments = fields.Text("Comments")

    @fields.depends('du')
    def on_change_with_operational_sector(self):
        if (self.du and self.du.operational_sector):
            return (self.du.operational_sector.id)



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011  Adrián Bernardi, Mario Puntin (health_invoice)
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                       HEALTH SERVICES LAB PACKAGE                     #
#                     health_services.py: Main module                   #
#########################################################################
from trytond.model import ModelView, fields
from trytond.pyson import Eval, Equal
from trytond.pool import Pool, PoolMeta
from trytond.i18n import gettext

from .exceptions import (NoServiceAssociated,
                         ServiceHasBeenUpdated)

__all__ = ['PatientLabTestRequest']


""" Add Lab order charges to service model """


class PatientLabTestRequest(metaclass=PoolMeta):
    'Lab Order'
    __name__ = 'gnuhealth.patient.lab.test'

    service = fields.Many2One(
        'gnuhealth.health_service', 'Service',
        domain=[('patient', '=', Eval('patient_id'))],
        depends=['patient_id'],
        states={'readonly': Equal(Eval('state'), 'done')},
        help="Service document associated to this Lab Request")

    service_updated = fields.Selection((
        ('yes', 'Yes'),
        ('no', 'No'),
        ('unknown', 'Unknown'),
    ), 'Service updated', sort=False)

    @classmethod
    def default_service_updated(self):
        return 'unknown'

    @classmethod
    def __setup__(cls):
        super(PatientLabTestRequest, cls).__setup__()
        cls._buttons.update({
            'update_service': {
                'readonly': Equal(Eval('state'), 'done'),
            },
        })

    @classmethod
    @ModelView.button
    def update_service(cls, laborders):
        laborder = laborders[0]

        if not laborder.service:
            raise NoServiceAssociated(
                gettext('health_services_lab.msg_no_service_associated'))

        if laborder.service_updated == 'yes':
            raise ServiceHasBeenUpdated(
                gettext('health_services_lab.msg_service_has_been_updated'))

        pool = Pool()
        HealthService = pool.get('gnuhealth.health_service')

        hservice = []
        service_data = {}
        service_lines = []

        # Add the laborder to the service document

        service_lines.append(('create', [{
            'product': laborder.name.product_id.id,
            'desc': laborder.name.product_id.rec_name,
            'qty': 1
        }]))

        hservice.append(laborder.service)

        description = "Services and Lab"

        service_data['desc'] = description
        service_data['service_line'] = service_lines

        HealthService.write(hservice, service_data)

        cls.write(laborders, {'service_updated': 'yes'})



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
import trytond.tests.test_tryton
from trytond.tests.test_tryton import ModuleTestCase


class HealthGeneticsUniprotTestCase(ModuleTestCase):
    '''
    Test HealthGeneticsUniprot module.
    '''
    module = 'health_genetics_uniprot'


del ModuleTestCase


def suite():
    suite = trytond.tests.test_tryton.suite()
    suite.addTests(unittest.TestLoader().loadTestsFromTestCase(
        HealthGeneticsUniprotTestCase))
    return suite



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                       HEALTH GENETICS package                         #
#                  health_genetics.py: main module                      #
#########################################################################
from trytond import backend
from trytond.model import ModelView, ModelSQL, fields, Unique
from trytond.pyson import Eval
from trytond.pool import Pool, PoolMeta
from uuid import uuid4
from trytond.modules.health.core import (get_institution,
                                         format_years_months_days)
from trytond.transaction import Transaction

__all__ = ['Gene', 'ProteinDisease', 'GeneVariant',
           'GeneVariantPhenotype',
           'PatientGeneticRisk', 'GnuHealthPatient']


class Gene(ModelSQL, ModelView):
    'Genes'
    __name__ = 'gnuhealth.gene'

    name = fields.Char(
        'Symbol', help='Symbol', required=True)

    aliases = fields.Char(
        'Aliases', help='Symbol aliases')

    name_aliases = fields.Char(
        'Name Aliases', help='Name aliases')

    hgnc_id = fields.Char(
        'HGNC ID', help='HUGO Gene Nomenclature Committee identifier',
        required=True)

    gene_type = fields.Selection([
        (None, ''),
        ('protein_coding', 'protein-coding gene'),
        ('ncrna_long_non_coding_rna', 'ncRNA: long non-coding RNA'),
        ('ncrna_y_rna', 'ncRNA: Y RNA'),
        ('ncrna_cluster_rna', 'ncRNA: cluster RNA'),
        ('ncrna_micro_rna', 'ncRNA: micro RNA'),
        ('ncrna_misc_rna', 'ncRNA: misc RNA'),
        ('ncrna_ribosomal_rna', 'ncRNA: ribosomal RNA'),
        ('ncrna_small_nuclear_rna', 'ncRNA: small nuclear RNA'),
        ('ncrna_small_nucleolar_rna', 'ncRNA: small nucleolar RNA'),
        ('ncrna_transfer_rna', 'ncRNA: transfer RNA'),
        ('ncrna_vault_rna', 'ncRNA: vault RNA'),
        ('pseudogene_pseudogene', 'pseudogene: pseudogene'),
        ('pseudogene_tcell_receptor',
            'pseudogene: T cell receptor pseudogene'),
        ('pseudogene_immunoglobulin', 'pseudogene: immunoglobulin pseudogene'),
        ('other_tcell_receptor_gene', 'other: T cell receptor gene'),
        ('other_complex_locus_constituent',
            'other: complex locus constituent'),
        ('other_endogenous_retrovirus', 'other: endogenous retrovirus'),
        ('other_fragile_site', 'other: fragile site'),
        ('other_immunoglobulin_gene', 'other: immunoglobulin gene'),
        ('other_readthrough', 'other: readthrough'),
        ('other_region', 'other: biological region'),
        ('other_virus_integration_site', 'other: virus integration site'),
        ('other_unknown', 'other: unknown'),
    ], 'Gene type', help="Locus in the form of group:type",
        sort=False)

    protein_name = fields.Char('Protein Code',
                               help="Encoding Protein Code,"
                               " such as UniProt protein name")

    # Do not translate the gene long name. Having the gene long name
    # description in English is OK in the scientific community, and it
    # will make the update process much faster, and don't overload the
    # translation server at Weblate. more details:
    # https://savannah.gnu.org/bugs/?64542
    long_name = fields.Char('Official Name', translate=False)
    gene_id = fields.Char('Entrez Gene ID',
                          help="Gene ID from NCBI Entrez database.")

    chromosome = fields.Char('Chromosome',
                             help="Name of the affected chromosome")

    location = fields.Char('Location', help="Locus of the chromosome")

    ensembl_id = fields.Char("Ensembl ID")
    refseq_accession = fields.Char("RefSeq", help="RefSeq Accession ID")
    omim_id = fields.Char("OMIM ID")

    info = fields.Text('Information', help="Extra Information")
    variants = fields.One2Many('gnuhealth.gene.variant', 'gene',
                               'Variants')

    protein_uri = fields.Function(fields.Char("Protein URI"),
                                  'get_protein_uri')

    def get_protein_uri(self, name):
        ret_url = ''
        if (self.protein_name):
            ret_url = 'http://www.uniprot.org/uniprot/' + \
                str(self.protein_name)
        return ret_url

    @classmethod
    def __setup__(cls):
        super(Gene, cls).__setup__()

        t = cls.__table__()
        cls._sql_constraints = [
            ('name_unique', Unique(t, t.hgnc_id),
                'The official identifier must be unique'),
        ]

    def get_rec_name(self, name):
        protein = ''
        if self.protein_name:
            protein = ' (' + self.protein_name + ') '
        return self.name + protein + ':' + self.long_name

    @classmethod
    def search_rec_name(cls, name, clause):
        if clause[1].startswith('!') or clause[1].startswith('not '):
            bool_op = 'AND'
        else:
            bool_op = 'OR'
        return [bool_op,
                ('name',) + tuple(clause[1:]),
                ('long_name',) + tuple(clause[1:]),
                ]

    @classmethod
    def __register__(cls, module):
        # Migration from 4.2:
        # rename gnuhealth.disease.gene to gnuhealth.gene
        backend.TableHandler.table_rename('gnuhealth_disease_gene', cls._table)

        # Update the data field from gnuhealth.disease.gene to gnuhealth.gene
        cursor = Transaction().connection.cursor()
        cursor.execute("""
            UPDATE ir_model_data SET model = 'gnuhealth.gene'
            WHERE model = 'gnuhealth.disease.gene'
            """)
        super().__register__(module)


class ProteinDisease(ModelSQL, ModelView):
    'Protein related disorders'
    __name__ = 'gnuhealth.protein.disease'

    name = fields.Char('Disease', required=True,
                       help="Uniprot Disease Code")

    disease_name = fields.Char('Disease name', translate=True)
    acronym = fields.Char('Mnemonic', required=True,
                          help="Disease acronym / mnemonics")

    disease_uri = fields.Function(fields.Char("Disease URI"),
                                  'get_disease_uri')

    mim_reference = fields.Char('MIM', help="MIM -"
                                "Mendelian Inheritance in Man- DB reference")

    gene_variant = fields.One2Many('gnuhealth.gene.variant.phenotype',
                                   'phenotype',
                                   'Natural Variant',
                                   help="Natural variants "
                                        "involved in this condition")

    keywords = fields.Char('Keywords')

    xrefs = fields.Char('Xrefs', help="Cross references")

    inheritance_pattern = fields.Selection([
        (None, ''),
        ('ad', 'Autosomal dominant'),
        ('ar', 'Autosomal recessive'),
        ('x', 'X-Linked'),
        ('y', 'Y-Linked'),
        ('m', 'Mitochondrial'),
        ('c', 'codominance'),
    ], 'Inheritance Pattern', help="Inheritance pattern",
        sort=False)

    description = fields.Text('Description', translate=True)

    active = fields.Boolean('Active', help="Whether this code is current."
                            "If you deactivate it, the code will "
                            "no longer show in the"
                            " protein-related diseases")

    @staticmethod
    def default_active():
        return True

    def get_disease_uri(self, name):
        ret_url = ''
        if (self.name):
            ret_url = 'http://www.uniprot.org/diseases/' + \
                str(self.name)
        return ret_url

    @classmethod
    def __setup__(cls):
        super(ProteinDisease, cls).__setup__()

        t = cls.__table__()
        cls._sql_constraints = [
            ('name_unique', Unique(t, t.name),
                'The Disease Code  name must be unique'),
        ]

    @classmethod
    def __register__(cls, module):
        # Migration from 4.2:
        # rename dominance field to inheritance_pattern
        table_h = cls.__table_handler__(module)
        table_h.column_rename('dominance', 'inheritance_pattern')
        super().__register__(module)

    def get_rec_name(self, name):
        return self.name + ':' + self.disease_name

    @classmethod
    def search_rec_name(cls, name, clause):
        if clause[1].startswith('!') or clause[1].startswith('not '):
            bool_op = 'AND'
        else:
            bool_op = 'OR'
        return [bool_op,
                ('name',) + tuple(clause[1:]),
                ('disease_name',) + tuple(clause[1:]),
                ]


class GeneVariant(ModelSQL, ModelView):
    'Natural Variant'
    __name__ = 'gnuhealth.gene.variant'

    gene = fields.Many2One('gnuhealth.gene', 'Gene',
                           required=True,
                           help="Gene and product (in parenthesis)")
    variant = fields.Char(
        "FTId", help="Variant Feature Identifier (FTId)",
        required=True)
    protein = fields.Char('Protein', help='Uniprot Protein ID')
    aa_change = fields.Char('AA Change', help="Amino acid change")

    dbsnp = fields.Char('dbSNP', help='dbSNP ID')

    dbsnp_url = fields.Function(
        fields.Char("Reference SNP",
                    help="Reference SNP (rs) link"), 'get_dbsnp_url')

    significance = fields.Selection([
        (None, ''),
        ('lbb', 'LB/B: Likely benign or benign'),
        ('lpp', 'LP/P: Likely pathogenic or pathogenic'),
        ('us', 'US: Unknown significance'),
    ], 'Significance',
        help="Category related to the clinical significance of the variant",
        sort=False)

    phenotypes = fields.One2Many('gnuhealth.gene.variant.phenotype', 'variant',
                                 'Phenotypes / Diseases')

    def get_dbsnp_url(self, name):
        url = ''
        if (self.dbsnp):
            url = f'https://www.ncbi.nlm.nih.gov/snp/{str(self.dbsnp)}'
        return url

    @classmethod
    def __setup__(cls):
        super(GeneVariant, cls).__setup__()

        t = cls.__table__()
        cls._sql_constraints = [
            ('variant_unique', Unique(t, t.variant),
                'The variant ID must be unique'),
            ('aa_unique', Unique(t, t.variant, t.aa_change),
                'The amino acid change for the variant already exists'),
        ]

    def get_rec_name(self, name):
        return ' : '.join([self.variant, self.aa_change])

    # Allow to search by gene and variant or amino acid change
    @classmethod
    def search_rec_name(cls, name, clause):
        if clause[1].startswith('!') or clause[1].startswith('not '):
            bool_op = 'AND'
        else:
            bool_op = 'OR'
        return [bool_op,
                ('gene',) + tuple(clause[1:]),
                ('variant',) + tuple(clause[1:]),
                ('aa_change',) + tuple(clause[1:]),
                ('dbsnp',) + tuple(clause[1:]),
                ]

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to patient
        if (table_h.column_exist('name')
                and not table_h.column_exist('gene')):
            table_h.column_rename('name', 'gene')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class GeneVariantPhenotype(ModelSQL, ModelView):
    'Variant Phenotypes'
    __name__ = 'gnuhealth.gene.variant.phenotype'

    name = fields.Char('Code', required=True)
    variant = fields.Many2One('gnuhealth.gene.variant', 'Variant',
                              required=True)

    gene = fields.Function(fields.Many2One(
        'gnuhealth.gene', 'Gene & Protein',
        depends=['variant'],
        help="Gene and expressing protein (in parenthesis)"),
        'get_gene',
        searcher='search_gene')

    phenotype = fields.Many2One('gnuhealth.protein.disease', 'Phenotype',
                                required=True)

    def get_gene(self, name):
        if (self.variant):
            return self.variant.gene.id

    def get_rec_name(self, name):
        if self.phenotype:
            return self.phenotype.rec_name

    @classmethod
    def search_gene(cls, name, clause):
        res = []
        value = clause[2]
        res.append(('variant.gene', clause[1], value))
        return res

    # Allow to search by gene, variant or phenotype
    @classmethod
    def search_rec_name(cls, name, clause):
        if clause[1].startswith('!') or clause[1].startswith('not '):
            bool_op = 'AND'
        else:
            bool_op = 'OR'
        return [bool_op,
                ('variant',) + tuple(clause[1:]),
                ('phenotype',) + tuple(clause[1:]),
                ('gene',) + tuple(clause[1:]),
                ]

    @classmethod
    def __setup__(cls):
        super(GeneVariantPhenotype, cls).__setup__()

        t = cls.__table__()
        cls._sql_constraints = [
            ('code', Unique(t, t.name),
                'This code already exists'),
        ]


class PatientGeneticRisk(ModelSQL, ModelView):
    'Patient Genetic Information'
    __name__ = 'gnuhealth.patient.genetic.risk'

    patient = fields.Many2One('gnuhealth.patient', 'Patient')
    disease_gene = fields.Many2One('gnuhealth.gene',
                                   'Gene', required=True)
    natural_variant = fields.Many2One('gnuhealth.gene.variant', 'Variant',
                                      domain=[('gene', '=',
                                              Eval('disease_gene'))],
                                      depends=['disease_gene'])

    zygosity = fields.Selection([
        (None, ''),
        ('homo', 'Homozygosity'),
        ('hetero', 'Heterozygosity'),
        ('hemi', 'Hemizygosity'),
    ], 'Zygosity', sort=False)

    variant_phenotype = fields.Many2One('gnuhealth.gene.variant.phenotype',
                                        'Phenotype',
                                        domain=[('variant', '=',
                                                Eval('natural_variant'))],
                                        depends=['natural_variant'])

    onset = fields.Integer('Onset', help="Age in years")

    notes = fields.Text("Notes")

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Health Prof',
        help="Health professional")

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    @staticmethod
    def default_institution():
        return get_institution()

    @classmethod
    def create_genetics_pol(cls, genetic_info):
        """ Adds an entry in the person Page of Life
            related to this genetic finding
        """
        Pol = Pool().get('gnuhealth.pol')
        pol = []

        vals = {
            'page': str(uuid4()),
            'person': genetic_info.patient.party.id,
            'age': format_years_months_days(
                years=genetic_info.onset, months=0, days=0),
            'federation_account':
                genetic_info.patient.party.federation_account,
            'page_type': 'medical',
            'medical_context': 'genetics',
            'relevance': 'important',
            'gene': genetic_info.disease_gene.rec_name,
            'natural_variant': genetic_info.natural_variant and
                               genetic_info.natural_variant.aa_change,
            'summary': genetic_info.notes,
            'author': genetic_info.healthprof and
            genetic_info.healthprof.party.rec_name,
            'node': genetic_info.institution and
            genetic_info.institution.party.name
        }
        if (genetic_info.variant_phenotype):
            vals['health_condition_text'] = vals['health_condition_text'] = \
                genetic_info.variant_phenotype.phenotype.rec_name

        pol.append(vals)
        Pol.create(pol)

    @classmethod
    def create(cls, vlist):

        # Execute first the creation of PoL
        genetic_info = super(PatientGeneticRisk, cls).create(vlist)

        cls.create_genetics_pol(genetic_info[0])

        return genetic_info

    @classmethod
    def search_rec_name(cls, name, clause):
        if clause[1].startswith('!') or clause[1].startswith('not '):
            bool_op = 'AND'
        else:
            bool_op = 'OR'
        return [bool_op,
                ('patient',) + tuple(clause[1:]),
                ('disease_gene',) + tuple(clause[1:]),
                ('variant_phenotype',) + tuple(clause[1:]),
                ]


class GnuHealthPatient (metaclass=PoolMeta):
    __name__ = 'gnuhealth.patient'

    """
    Add to the Medical patient_data class (gnuhealth.patient) the genetic
    and family risks
    """

    genetic_risks = fields.One2Many('gnuhealth.patient.genetic.risk',
                                    'patient', 'Genetic Information')
